﻿#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3				// Use modern global access method and strict wave access
#pragma DefaultTab={3,20,4}		// Set default tab width in Igor Pro 9 and later
#pragma version = 1.00

// NOTE: not tested! Use at own risk!

function CheckProcVersions()
	
	wave /T wIncludedProcs = ListToTextWave(WinList("*", ";", "INCLUDE:2"), ";")
	wave /T wProcFilePaths = getProcsRecursive(1)
	duplicate /free/T wProcFilePaths wProcFileNames
	wProcFileNames = ParseFilePath(0, wProcFilePaths, ":", 1, 0)
	string strWarning = "", loadedPath=""
	int index, numDuplicates
	variable fileVersion, loadedVersion
	numDuplicates = 0
	for (string procedure : wIncludedProcs)
		index = -1
		do
			FindValue /S=(index + 1)/TEXT=procedure/TXOP=4 wProcFileNames
			if (v_row < 0)
				break
			endif
			index = v_row
			loadedPath = getProcWinFilePath(procedure)
			if (cmpstr(loadedPath, wProcFilePaths[index]) == 0)
				continue
			endif
			numDuplicates += 1
			loadedVersion = GetProcedureVersion(procedure)
			fileVersion = getProcVersion(wProcFilePaths[index])
			sprintf strWarning "A %s of the included file\r\r%s\r\rexists at\r\r%s", selectstring(fileVersion > loadedVersion, "duplicate", "more recent version"), loadedPath, wProcFilePaths[index]
			doAlert 0, strWarning // no check for too long string
		while (index < numpnts(wProcFileNames)-1)
	endfor
	if (numDuplicates == 0)
		doAlert 0, "No Duplicates Found"
	endif
end

// based on https://www.wavemetrics.com/code-snippet/determine-version-procedure-file
Function GetProcedureVersion(procedureWinTitleStr)
    String procedureWinTitleStr
   
    // By default, all procedures are version 0 unless
    // otherwise specified.
    Variable version = 0
    Variable versionIfError = NaN
   
    String procText = ProcedureText("", 0, procedureWinTitleStr)
    if (strlen(procText) <= 0)
        return versionIfError       // Procedure window doesn't exist.
    endif
   
    String regExp = "(?i)(?:^#pragma|\\r#pragma)(?:[ \\t]+)version(?:[\ \t]*)=(?:[\ \t]*)([\\d.]*)"
   
    String versionFoundStr
    SplitString/E=regExp procText, versionFoundStr
    if (V_flag == 1)
        version = str2num(versionFoundStr)
    endif
    return version 
End

// ------------------- from Updater ---------------

// search recursively for ipf files and return list of files as textwave
function /wave getProcsRecursive(int fast)
	
	int maxFolders = 1000, fullpath = 1
	int folderIndex, i
	string strFile = "", strFolder = ""
	Make /free/T/N=0 w_folders, w_allFiles
	
	w_folders = {SpecialDirPath("Igor Pro User Files",0,0,0) + "User Procedures:"} // search recursively starting in this folder

	// w_folders will grow as subfolders are added
	for (folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
		// check files in folderIndexth folder from w_folders
		NewPath /O/Q/Z tempPathIXI, w_folders[folderIndex]
		
		if (fast) // don't worry about semicolons in filenames, or shortcuts ending in .ipf!
			// add list of ipf files in current folder to w_allFiles
			wave /T w = ListToTextWave(IndexedFile(tempPathIXI, -1, ".ipf"),";")
			if (fullpath)
				w = w_folders[folderIndex] + w
			endif
			TextWaveConcatenate(w_allFiles, w)
			// add list of folders in current folder to w_folders
			wave /T w = ListToTextWave(IndexedDir(tempPathIXI, -1, 1),";")
			w += ":"
			TextWaveConcatenate(w_folders, w)
			
#ifdef WINDOWS
			// get a list of shortcuts
			wave /T w = ListToTextWave(IndexedFile(tempPathIXI, -1, ".lnk"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				Duplicate /free w w_f
				// keep shortcuts to files in w, shortcuts to folders in w_f
				if (fullpath == 0)
					w = ParseFilePath(0, w, ":", 1, 0)
				endif
				TextWaveStringMatch(w, "*.ipf")
				TextWaveConcatenate(w_allFiles, w)
				TextWaveStringMatch(w_f, "*:")
				TextWaveConcatenate(w_folders, w_f)
			endif
#else // mac
			// get a list of file aliases
			wave /T w = ListToTextWave(IndexedFile(tempPathIXI, -1, "alis"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				if (fullpath == 0)
					w = ParseFilePath(0, w, ":", 1, 0)
				endif
				TextWaveStringMatch(w, "*.ipf")
				TextWaveConcatenate(w_allFiles, w)
			endif
			// get a list of folder aliases
			wave /T w = ListToTextWave(IndexedFile(tempPathIXI, -1, "fdrp"),";")
			if (numpnts(w))
				w = ResolveAlias(w_folders[folderIndex] + w)
				TextWaveZapString(w, "")
				TextWaveConcatenate(w_folders, w)
			endif
#endif
			continue // next folder
		endif
				
		// the slow way: loop through files in folder
		for (i=0;1;i+=1)
			strFile = IndexedFile(tempPathIXI, i, "????")
			if (strlen(strFile) == 0)
				break
			endif
			GetFileFolderInfo /Q/Z w_folders[folderIndex] + strFile
			if (V_isAliasShortcut)
				strFile = ResolveAlias(w_folders[folderIndex] + strFile)
				if (stringmatch(strFile,"*:"))
					w_folders[numpnts(w_folders)] = {strFile}
				endif
				if (stringmatch(strFile,"*.ipf"))
					if (fullpath == 0)
						strFile = ParseFilePath(0, strFile, ":", 1, 0)
					endif
					w_allFiles[numpnts(w_allFiles)] = {strFile}
				endif
				continue // next file
			endif
			
			if (stringmatch(strFile,"*.ipf"))
				if (fullpath)
					strFile = w_folders[folderIndex] + strFile
				endif
				w_allFiles[numpnts(w_allFiles)] = {strFile}
				continue
			endif
		endfor // next file
		
		// add subfolders in current folder
		for (i=0;1;i+=1)
			strFolder = IndexedDir(tempPathIXI, i, 1)
			if (strlen(strFolder) == 0)
				break
			endif
			w_folders[numpnts(w_folders)] = {strFolder + ":"}
		endfor // next subfolder
		if (folderIndex == maxFolders)
			Print "Warning: getProcsRecursive exceeding max iterations"
			break
		endif
	endfor // next folder
	
	KillPath /Z tempPathIXI
	return w_allFiles
end

// appends w2 to w
function TextWaveConcatenate(wave /T w, wave /T w2)
	int oldPnts = numpnts(w), newPnts = numpnts(w2)
	if (newPnts)
		Redimension /N=(oldPnts+newPnts), w
		w[oldPnts,] = w2[p-oldPnts]
	endif
end

// Recursively resolve shortcuts to files/directories
// returns full path or an empty string if the file does not exist or the
// shortcut points to a non existing file/folder
function/S ResolveAlias(string strPath) // full path to file, folder or alias
	GetFileFolderInfo/Q/Z RemoveEnding(strPath, ":")
	if (V_isAliasShortcut)
		return ResolveAlias(S_aliasPath)
	endif
	if (v_flag == 0)
		return strPath // path to a file or folder
	endif
	return ""
end

// keep only points of textwave matching strMatch
function TextWaveStringMatch(wave /T w, string strMatch)
	if (numpnts(w))
		w = SelectString(stringmatch(w,strMatch),"",w)
		TextWaveZapString(w, "")
	endif
end

// Remove points from textwave w (and optionally the corresponding points
// of waves in wave reference wave w_zapRefs) that match matchStr. Does
// not check numbers of points in w_zapRefs waves. Wildcards okay.
function TextWaveZapString(wave /T w, string matchStr, [wave /Z/WAVE w_zapRefs])
	int i, j
	for (i=numpnts(w)-1;i>=0;i-=1)	
		if (stringmatch(w[i],matchStr))
			DeletePoints i, 1, w
			if (WaveExists(w_zapRefs))
				for (j=0;j<numpnts(w_zapRefs);j+=1)
					DeletePoints i, 1, w_zapRefs[j]
				endfor
			endif
		endif
	endfor
	return numpnts(w)
end

// extract procedure version from file
function getProcVersion(string filePath)
	variable procVersion
	Grep /Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z filePath
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	if (V_flag!=1 || procVersion<=0)
		return 0
	endif
	return procVersion
end

function /S getProcWinFilePath(string strWin)
	GetWindow /Z $strWin file
	return StringFromList(1, S_value) + StringFromList(0, S_value)
end